from fastapi import APIRouter, Depends, UploadFile, HTTPException, Response, Request
from starlette_context import context
import logging
from requests.structures import CaseInsensitiveDict
from auth import auth
import opa
import correlation
from bundles import bundle
from string import Template
import hashlib

from pydantic import BaseModel

from starlette.status import (
    HTTP_400_BAD_REQUEST,
    HTTP_422_UNPROCESSABLE_ENTITY,
)

router = APIRouter()
logger = logging.getLogger(__name__)


class Detail(BaseModel):
    detail: str


@router.put(
    "/validate/{policy_id}",
    responses={
        HTTP_400_BAD_REQUEST: {"model": Detail},
        HTTP_422_UNPROCESSABLE_ENTITY: {"model": Detail},
    },
)
def validate_policy(
    response: Response,
    request: Request,
    policy_id: str,
    file: UploadFile,
    template: bool = False,
    auth_data: auth.Auth = Depends(auth.require_authorized_admin),
):
    """
    # Validate Policy
    This API checks to make sure the rego is valid and the naming of the policy package is acceptable.

    If template parameter is True, then the incoming file will automatically replace the following during validation:
    - data_partition
    - DATA_PARTITION
    - name with policy_id without ".rego" suffix
    """
    logging.setLogRecordFactory(
        correlation.set_correlation_id(context["correlation_id"])
    )
    logger = logging.getLogger(__name__)
    response.headers["X-Correlation-ID"] = context["correlation_id"]

    if not policy_id.endswith(".rego"):
        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY,
            detail="Policy ID should end with a .rego",
        )

    contents = file.file.read()

    if template:
        template_data = Template(contents.decode("utf-8"))
        data = template_data.substitute(
            {
                "data_partition": auth_data.data_partition_id,
                "DATA_PARTITION": auth_data.data_partition_id,
                "name": policy_id.removesuffix(".rego"),
            }
        )
        contents = data.encode("utf-8")

    if not bundle.package_name_ok(auth_data.data_partition_id, policy_id, contents):
        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY,
            detail="Package declaration should be: {0}".format(
                bundle.correct_package_declaration(
                    auth_data.data_partition_id, policy_id
                )
            ),
        )

    data = contents.decode("utf-8")

    short_name = policy_id.removesuffix(".rego")

    if verify_policy_with_opa(
        policy_id=policy_id,
        auth_data=auth_data,
        data=data,
        correlation_id=context["correlation_id"],
    ):
        sha256 = hashlib.sha256(data.encode()).hexdigest()
        response.headers["X-SHA-256"] = sha256
        return f'{policy_id} valid as osdu.partition["{auth_data.data_partition_id}"].{short_name} {sha256}'
    else:
        # Not expected to get to this code block
        logger.error(f"Invalid {policy_id}, unexpected return from OPA")
        raise HTTPException(status_code=HTTP_400_BAD_REQUEST, detail="Invalid request")


def verify_policy_with_opa(
    policy_id: str, auth_data: auth.Auth, data: str, correlation_id: str
):
    if not policy_id.endswith(".rego"):
        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY,
            detail="Policy ID should end with a .rego",
        )

    data_partition = auth_data.data_partition_id

    headers = CaseInsensitiveDict()
    headers["X-Correlation-ID"] = correlation_id
    headers["req_id"] = correlation_id
    headers["Authorization"] = "Bearer " + auth_data.access_token
    headers["data-partition-id"] = data_partition
    headers["Content-Type"] = "application/octet-stream"

    data = data.replace("package osdu.partition[", "package tmp.osdu.partition[")

    temp_id = f"tmp/{data_partition}/{policy_id}"
    logger.info(f"TMP-AUDIT-OPA put id: {temp_id} user: {auth_data.user_id}")

    temp_result = opa.put_opa_policy_direct(
        policy_id=temp_id, data=data.encode("utf-8"), headers=headers, validate=True
    )
    if not temp_result.ok:
        logger.error(
            f"Error when talking to OPA to check syntax of policy {temp_result.message}"
        )
        raise HTTPException(
            status_code=temp_result.status_code,
            detail=f"Unable to validate policy! Error: {temp_result.message}.",
        )
    else:
        temp_result = opa.delete_opa_policy_direct(policy_id=temp_id)
        if not temp_result.ok:
            logger.warning(
                f"Error when talking to OPA to delete temp policy {temp_id} {temp_result.message}"
            )
            # raise HTTPException(status_code=temp_result.status_code,
            #    detail=f"OPA Error while deleting temp policy: {temp_result.message}."
            #    )
        else:
            logger.info(f"TMP-AUDIT-OPA delete id: {temp_id} user: {auth_data.user_id}")
        return True
