# translate/convert.py
from starlette_context import context
from operator import attrgetter

import requests
import logging
from elasticsearch_dsl import Search
from elasticsearch_dsl.query import (
    Q,
    Term,
    Range,
    Bool,
    SimpleQueryString,
    Regexp,
    Terms,
)

#from cachetools import cached, TTLCache
import correlation
from . import ast
import opa
from .translate_exceptions import PolicyDenyError
#import conf

logger = logging.getLogger(__name__)

def convert_opa_ast_json_to_es_request_body(opa_ast, opa_request_input):
    opa_queryset = ast.QuerySet.from_data(opa_ast)
    es_queries = []
    for opa_query in opa_queryset.queries:
        es_query = _convert_opa_query_to_es_query(opa_query, opa_request_input)
        es_queries.append(es_query)
    postprocessed_es_queries = _postprocess_es_queries(es_queries)
    return Search().query("bool", should=postprocessed_es_queries).to_dict()

def _convert_opa_query_to_es_query(opa_query: ast.Query, opa_request_input):
    """
    Return None if there are no conditions that could make the query true
    """
    logging.setLogRecordFactory(correlation.set_correlation_id(context["correlation_id"]))
    logger = logging.getLogger(__name__)
    es_and_filters = []
    side_effect_context = {}
    for expr in opa_query.exprs:
        logger.debug(f"convert opa query to es, evaluating expr {expr}")
        if expr.is_call():
            assert len(expr.operands) == 2, f"assert01 expected expr {expr} operands == 2"

            # Generate ES query only for those with OPA unknown
            # If it does not have the unknown, still evaluate it anyway for side effects
            has_unknown_term = None
            for operand in expr.operands:
                if has_unknown_term:  # Done searching through operands
                    break
                if isinstance(operand.value, ast.Ref):
                    ref_terms = operand.value.terms
                    # FIXME hardcoded OPA unknown
                    a = ref_terms[0].value.value == "input"
                    b = ref_terms[1].value.value == "record"
                    has_unknown_term = a and b
            if not has_unknown_term:
                logger.debug(f"performing side effect for {expr}")
                if expr.op() == "http.send":
                    logger.debug(f"isistance {expr.terms[2].value}, {ast.Var}")
                    assert isinstance(expr.terms[2].value, ast.Var), "assert02 isinstance"
                    logger.debug("after assert")
                    http_result = _resolve_http(expr)
                    side_effect_context[expr.terms[2].value.value] = http_result
                    logger.debug(f"side effect context updated: {side_effect_context}")
                elif expr.op() == "eq":
                    logger.debug(f"eq {expr.terms[1].value}, {str(ast.Ref)}")
                    assert isinstance(expr.terms[1].value, ast.Ref), "assert03 isinstance"
                    logger.debug("after assert eq 1")
                    the_ref = expr.terms[1]
                    logger.debug(f"the_ref: {the_ref}")
                    logger.debug(f"assert: {the_ref.value.terms[0].value} {ast.Var}")
                    assert isinstance(the_ref.value.terms[0].value, ast.Var), "assert04 isinstance"
                    logger.debug("after assert eq 2")
                    variable_key = the_ref.value.terms[0].value.value
                    logger.debug(f"variable_key: {variable_key}")
                    assert isinstance(the_ref.value.terms[1].value, ast.Scalar), "assert05 isinstance"
                    logger.debug("after assert eq 3")
                    variable_path = the_ref.value.terms[1].value.value
                    logger.debug(f"variable_path: {variable_path}")
                    logger.debug(f"side_effect_context: {side_effect_context}")
                    if variable_key in side_effect_context:
                        variable_value = side_effect_context[variable_key][variable_path]
                        logger.debug(f"variable_value: {variable_value}")
                    else:
                        logger.error(f"Unsupported evaluation {expr}")
                        if "deny" in str(expr):
                            raise PolicyDenyError(f"Unsupported Deny evaluation {expr}")
                        else:
                            raise NotImplementedError(f"Unsupported evaluation {expr}")
                    expected_variable_value = expr.terms[2].value.value
                    if variable_value == expected_variable_value:
                        logger.debug("eq check passed")
                    else:
                        logger.debug("eq check did not pass")
                        return None
                else:
                    logger.error(f"Unsupported side effect evaluation {expr}")
                    raise NotImplementedError(
                        "Unsupported side effect evaluation", str(expr)
                    )
                continue

            field_name = None
            field_value = None
            for operand in expr.operands:
                if isinstance(operand.value, ast.Scalar):
                    scalar = operand.value
                    field_value = scalar.value
                elif isinstance(operand.value, ast.Object):
                    assert False, "assert06 False"
                    logger.debug(operand.value)
                elif isinstance(operand.value, ast.Var):
                    assert False, "assert07 False"
                    logger.debug(operand.value)
                else:
                    not_scalar = operand.value
                    field_name = ".".join(
                        term.value.value
                        for term in not_scalar.terms[2:]
                        if not isinstance(term.value, ast.Var)
                    )
            assert field_name is not None, "assert08 field_name"
            assert field_value is not None, "assert09 field_value"

            if expr.op() == "eq":
                es_filter = Term(**{field_name: field_value})
            elif expr.op() == "gt":
                es_filter = Range(**{field_name: {"gt": field_value}})
            elif expr.op() == "lt":
                es_filter = Range(**{field_name: {"lt": field_value}})
            elif expr.op() == "gte":
                es_filter = Range(**{field_name: {"gte": field_value}})
            elif expr.op() == "lte":
                es_filter = Range(**{field_name: {"lte": field_value}})
            elif expr.op() == "neq":
                es_filter = Bool(must_not=Term(**{field_name: field_value}))
            elif expr.op() == "contains":
                es_filter = SimpleQueryString(
                    query=field_value, fields=[field_name], default_operator="AND"
                )
            elif expr.op() == "regex.match":
                es_filter = Regexp(**{field_name: {"value": field_value}})
            else:
                raise NotImplementedError("Unsupported OPA operator: " + expr.op())
            es_and_filters.append(es_filter)
        elif expr.is_every():
            (_, var_in_body, domain, body) = attrgetter(
                "key", "value", "domain", "body"
            )(expr.terms.value)

            # Validate domain is input.record.xxx
            assert domain.value.operand(0).value.value == "input", "assert10 input"
            assert domain.value.operand(1).value.value == "record", "assert11 record"

            # Support only simple body
            assert len(body) == 1, f"assert12 body len {body}"
            body = body[0]

            # Ensure body is negated
            # ElasticSearch, being an inverted index, cannot support universal quantification
            assert body.negated, "assert13 body negated"

            # TODO: Assume body is always a Ref
            assert isinstance( body.terms.value, ast.Ref), "assert14 isinstance"
            values = _resolve_ref(
                body.terms.value, var_in_body.value.value, opa_request_input
            )

            es_filter = Bool(
                must_not=Terms(**{(domain.value.operand(2).value.value): values})
            )
            es_and_filters.append(es_filter)
        else:
            raise NotImplementedError("Unsupported expr: " + expr)
    return Q("bool", filter=es_and_filters)


def _resolve_ref(ref, expected_variable_name, opa_request_input):
    logging.setLogRecordFactory(correlation.set_correlation_id(context["correlation_id"]))
    logger = logging.getLogger(__name__)
    logger.debug(f"resolve ref: {ref}")
    assert isinstance(ref.operand(0), ast.Term), "assert15 isinstance"
    assert isinstance(ref.operand(0).value, ast.Var), "assert16 isinstance"
    assert ref.operand(0).value.value == "data", "assert17 data"
    rego_package_name = ref.operand(1).value.value
    rego_rule_name = ref.operand(2).value.value
    variable_name = ref.operand(3).value.value
    assert variable_name == expected_variable_name, f"assert18 {variable_name} == {expected_variable_name}"
    return opa.get_document_with_input(
        rego_package_name, rego_rule_name, opa_request_input
    )


def _resolve_http(expr):
    logging.setLogRecordFactory(correlation.set_correlation_id(context["correlation_id"]))
    logger = logging.getLogger(__name__)
    logger.debug(f"resolve http: {expr}")
    assert isinstance(expr.terms[1].value, ast.Object), "assert19 isinstance"
    method = None
    url = None
    headers = {}
    for k, v in expr.terms[1].value.pairs:
        key = k.value.value
        val = v.value
        if key == "method":
            method = val.value
        elif key == "url":
            url = val.value
        elif key == "headers":
            for pair in val.pairs:
                header_name = pair[0].value.value
                header_value = pair[1].value.value
                headers[header_name] = header_value
        elif key == "raise_error":
            pass  # no-op
        else:
            raise NotImplementedError("Unsupported http.send option", str(key))
    assert method is not None, "assert20 method is None"
    assert url is not None, "assert21 url is None"
    logger.debug(f"http request method: {method}, url: {url}, headers: {headers}")
    resp = requests.request(method, url=url, headers=headers)
    response = {
        "status_code": resp.status_code,
        "body": resp.json(),
    }
    logger.debug(f"http response: {response}")
    return response


def _postprocess_es_queries(es_queries):
    logging.setLogRecordFactory(correlation.set_correlation_id(context["correlation_id"]))
    logger = logging.getLogger(__name__)
    logger.debug(f"es_queries: {es_queries}")

    def combine_term_queries(es_queries):
        """Iterate over the queries and combine term queries to terms query"""
        #logger = correlation.DefaultExtrasAdapter(_logger, {"correlation_id": context["correlation_id"]})

        # 1. Collect the term queries
        tqs = {}  # Collected term queries, keyed by ES field name
        for idx, es_query in enumerate(es_queries):
            if isinstance(es_query, Bool) and hasattr(es_query, "filter"):
                if len(es_query.filter) != 1:
                    # If there are multiple clauses, we cannot extract the
                    # term clause because the clauses are AND'ed. Skip it.
                    continue
                for filter in es_query.filter:
                    if isinstance(filter, Term):
                        (es_field_name, es_field_value) = next(
                            iter(filter._params.items())
                        )
                        if es_field_name in tqs:
                            tqs[es_field_name].append((idx, es_field_value))
                        else:
                            tqs[es_field_name] = [(idx, es_field_value)]
                    else:
                        pass  # no op, only looking for Term queries
            else:
                raise NotImplementedError(
                    "Unexpected ES query type in post processing step"
                )
        logger.debug(f"tqs: {tqs}")

        # 2. Generate terms query for each of the collected term queries
        terms_queries = []
        index_of_term_queries_to_delete = []
        for (es_field_name, collected_terms) in tqs.items():
            if len(collected_terms) > 1:
                terms = [item[1] for item in collected_terms]
                terms_query = Terms(**{es_field_name: terms})
                terms_queries.append(terms_query)
                idxs = [item[0] for item in collected_terms]
                index_of_term_queries_to_delete.extend(idxs)
            else:
                pass  # skip rewriting if it is just a single term
        logger.debug(f"terms_queries: {terms_queries}")
        logger.debug(f"index_of_term_queries_to_delete: {index_of_term_queries_to_delete}")

        # 3. Create the new list of es_queries
        new_es_queries = [
            q
            for i, q in enumerate(es_queries)
            if i not in index_of_term_queries_to_delete
        ]
        for terms_query in terms_queries:
            new_es_query = Q("bool", filter=terms_query)
            new_es_queries.append(new_es_query)
        logger.debug(f"new_es_queries: {new_es_queries}")
        return new_es_queries

    new_es_queries = combine_term_queries(es_queries)
    return new_es_queries
